unit as400;

// Paul TOTH <tothpaul@mygale.org>
(*
Copyright (C) 1998 Paul Toth (TothPaul@Mygale.org)
 http://www.mygale.org/~tothpaul

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

{
 Now support Delphi 3+ - {$IFNDEF VER90}
}

interface

uses
 windows,dialogs,forms,classes,stdctrls,sysutils,cawin32,EBCDIC
 {$IFNDEF VER90},Controls{$ENDIF};

Type
 {
  The Custom400 component has an Error property, so each API can be called by
   Error:=API(Parms);
  The SetError call OnError Event or Call the virtual InternalError procedure

  new : fNotifications, allow to link a list of methode to an event !
 }
  TNotify=(tnLoaded,tnOpen,tnClose,tnDestroy);
  TNotificationEvent=procedure(Sender:TObject; Notify:TNotify) of object;

  TCustom400=class(TComponent)
  private
   fError:integer;
   fNotifications:TList;
   EOnError:TNotifyEvent;
   Procedure SetError(Value:Integer);
  Protected
   Procedure InternalError; virtual;
  Public
   Property Error:integer read fError write SetError;
   Constructor Create(AOwner:TComponent); override;
   Destructor Destroy; override;
   procedure AddNotification(Notify:TNotificationEvent);
   procedure RemoveNotification(Notify:TNotificationEvent);
   procedure DoNotifications(Notify:TNotify);
  Published
   Property OnError:TNotifyEvent read EOnError write EOnError;
  end;

 { Error Monitor -- not used yet ... }
  TErrorMonitor=class(TCustom400)
  private
   fHostMsgId:THandle;
   fFirstLevelMsg:THandle;
   fSecondLevelMsg:THandle;
  //protected--no !
  public
   Constructor Create(AOwner:TComponent); override;
   Destructor Destroy; override;
  end;

 { AS400 provide a link to an AS/400 :) }

  TAS400 = class(TCustom400)
  private
    fDBHandle:THandle; // data base
    fDBHCount:integer;
    fRCHandle:THandle; // remote command
    fRCHCount:integer;
    fSysName:string;    // as displayed in IDE
    fSysUsed:string;    // fSysName or Retrieved sysname for *Default
    Procedure SetSysName(Value:string);
    Procedure SetNoWrite(Value:string);
    Function GetUserName:string;
    Function GetVersion:string;
  protected
    Procedure Loaded; override;
  public
    Constructor Create(AOwner:TComponent); override;
    Destructor Destroy; override;
    Procedure GetLibList(List:TStrings);
    Procedure GetLibFiles(LibName:string; FileSelection,FileAttributes:dword; List:TStrings);
    Procedure CreateDBHandle;
    Procedure DeleteDBHandle;
    function DBHandle:THandle;
    Procedure CreateRCHandle;
    Procedure DeleteRCHandle;
    function RCHandle:THandle;
  published
    Property  SysName:string read fSysname write SetSysName;
    Property  UserName:string read GetUserName write SetNoWrite stored false;
    Property  Version:string read GetVersion write SetNoWrite stored false;
  end;

  { TSysList }
  TSysList=class(TCustomComboBox)
  private
   fAutoSelect:boolean; // select first AS/400 on startup
   fAS400:TAS400;       // change AS400.SysName on selection
   procedure SetAutoSelect(Value:boolean);
   procedure SetAS400(Value:TAS400);
   procedure setsysname(Value:string); // tired of using Items[ItemIndex]
   function getsysname:string;
  protected
   procedure Change; override;
   procedure UpdateAS400; // Set AS400.SysName (disconnect & reconnect)
  public
   procedure Loaded; override;
   procedure Notification(AComponent: TComponent; Operation: TOperation); override;
   property Sysname:string read getsysname write setsysname;
  published
   property AutoSelect:boolean read fAutoSelect write SetAutoSelect;
   property AS400:TAS400 read fAS400 write SetAS400;
   property Sorted;
   property OnChange;
  end;

  { TLibFiles }
  TSelectFiles=(sfAll,sfSource,sfData);
  TFileAttributes=(faAll,faPhysical,faLogical,faODBCTable,faODBCView);
  TLibFiles=class(TCustomListBox)
  private
   fAS400:TAS400;
   fLibName:string;
   fSelectFiles:TSelectFiles;
   fFileAttributes:TFileAttributes;
   fActive:boolean; // use this to avoid refresh when setting attributes
   procedure SetAS400(Value:TAS400); // add/remove notification
   procedure AS400Notify(Sender:TObject; Notify:TNotify); // do the job :)
   procedure SetLibName(Value:string);
   procedure SetSelectFiles(Value:TSelectFiles);
   procedure SetFileAttributes(Value:TFileAttributes);
   procedure reload;
  public
   constructor Create(AOwner:TComponent); override;
   destructor Destroy; override; // notify AS/400 when destroyed !
  published
   property AS400:TAS400 read fAS400 write SetAS400;
   property LibName:string read fLibName write SetLibName;
   property SelectFiles:TSelectFiles read fSelectFiles write SetSelectFiles;
   property FileAttributes:TFileAttributes read fFileAttributes write SetFileAttributes;
   Property Active:boolean read fActive write fActive;
   property Sorted;
   property OnClick;
  end;

  { TLibFileSelection }
  TLibFileSelection=class(TCustomComboBox)
  private
   fLibFiles:TLibFiles;
   procedure SetLibFiles(Value:TLibFiles);
  protected
   procedure Change; override;
   procedure UpdateLibFiles;
  public
   procedure Loaded; override;
   procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  published
   property LibFiles:TLibFiles read fLibFiles write SetLibFiles;
  end;

{ TLibList }
  TLibList=class(TCustomListBox)
  private
   fAS400:TAS400;
   fLibFiles:TLibFiles;
   fAutoSelect:boolean; // select first Lib on startup
   procedure SetAS400(Value:TAS400); // add/remove notification
   procedure SetLibFiles(Value:TLibFiles);
   procedure AS400Notify(Sender:TObject; Notify:TNotify); // do the job :)
   procedure SetAutoSelect(Value:boolean);
  protected
   procedure Click; override;
   procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
   destructor Destroy; override; // notify AS/400 when destroyed !
  published
   property AS400:TAS400 read fAS400 write SetAS400;
   property LibFiles:TLibFiles read fLibFiles write SetLibFiles;
   property AutoSelect:boolean read fAutoSelect write SetAutoSelect;
   property Sorted;
   property OnClick;
  end;

//--- Call AS/400 pgm ----
  TCall400=class;

  TParmType=(tChar,tDec,tBin);

  TParm400=class(TComponent)
   private
    fCall400:TCall400;
    fParmType:TParmType; // 0=*CHAR, 1=*DEC, 2=*BIN
    fSize:word;
    fPrecision:word;
    fBind:pointer;
    procedure SetSize(Value:word);
    Function GetString:string;
    Procedure SetString(Value:string);
   protected
    procedure EbcdicToAnsi;
    procedure AnsiToEbcdic;
   public
    function HasParent: Boolean; override; // "override" parent
    function GetParentComponent: TComponent; override; // return "new" parent
    procedure SetParentComponent(AParent: TComponent); override; // needed by the Editor
    procedure ReadState(Reader: TReader); override;
    Constructor Create(AOwner:TComponent); override;
    Destructor Destroy; override;
    Procedure Add;
    procedure Clear;
    Property Bind:pointer read fBind;
    Property AsString:string read GetString write SetString stored false;
   published
    Property ParmType:TParmType read fParmType write fParmType;
    Property Size:word read fSize write SetSize;
    Property Precision:word read fPrecision write fPrecision;
  end;

 TCall400=class(TCustom400)
  private
    fAS400:TAS400;
    fActive:boolean;
    fActivate:boolean;
    fPgmName:string;
    fLibName:string;
    fPgmHandle:THandle;
    fParms:TList;
    procedure SetAS400(Value:TAs400);
    function GetParm(Index:Integer):TParm400;
    function GetParmCount:integer;
    procedure SetActive(Value:boolean);
  protected
    procedure Activate;
    procedure Notify400(Sender:TObject; Notify:TNotify);
   {$IFDEF VER90}
    procedure GetChildren(Proc: TGetChildProc); override;
   {$ELSE}
    procedure GetChildren(Proc: TGetChildProc;Root:TComponent); override;
   {$ENDIF}
  public
    Constructor Create(AOwner:TComponent); override;
    Destructor Destroy; override;
    Property Parms[Index:integer]:TParm400 read GetParm stored true;
    Property ParmCount:integer read GetParmCount;
    Function ParmByName(Const ParmName:string):TParm400;
    Function AddParm(AName:String; AType:TParmType; ASize,APrecision:word):pointer;
    Function Execute:boolean;
    Procedure AssignParms(ACall400:TCall400);
  published
    Property AS400:TAS400 read fAS400 write SetAS400;
    Property PgmName:string read fPgmName write fPgmName;
    Property LibName:string read fLibName write fLibName;
    Property Active:boolean read fActive write SetActive;
  end;


Procedure GetSysList(List:TStrings); // this function do not need TAS400

implementation

{ For now, I'll use a single ErrorHandle ...wait and see... }
Var
 ErrHandle:THandle; // error Handle
 ErrorMonitor:TErrorMonitor;

{ Constants --------------------------------------------------------- }
Const
 cwbTitle      ='VCL/400';
 cwbNoErrHandle='Impossible d''initialiser la gestion des erreurs';
 cwbNotWhileActive='Ne peut effectuer cette opration sur un composant actif';
 cwbUnknowSystem  ='systme inconnu';
 cwbParmNotFound  ='Paramtre non trouv';

{ TErrorMonotir ----------------------------------------------------- }
Constructor TErrorMonitor.Create(AOwner:TComponent);
 begin
  Inherited Create(AOwner);
  Error:=cwbDB_CreateDataHandle(fHostMsgID, ErrHandle);
  Error:=cwbDB_CreateDataHandle(fFirstLevelMsg, ErrHandle);
  Error:=cwbDB_CreateDataHandle(fSecondLevelMsg, ErrHandle);
 end;

Destructor TErrorMonitor.Destroy;
 begin
  Error:=cwbDB_DeleteDataHandle(fSecondLevelMsg, ErrHandle);
  Error:=cwbDB_DeleteDataHandle(fFirstLevelMsg, ErrHandle);
  Error:=cwbDB_DeleteDataHandle(fHostMsgID, ErrHandle);
  inherited Destroy;
 end;

{ TCustom400 -------------------------------------------------------- }
Constructor TCustom400.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  fNotifications:=TList.Create;
 end;

 Destructor TCustom400.Destroy;
  begin
   DoNotifications(tnDestroy);
   fNotifications.Free;
   inherited Destroy;
  end;

 procedure TCustom400.AddNotification(Notify:TNotificationEvent);
  var
   LocalCopy:^TNotificationEvent;
  begin
   New(LocalCopy);
   LocalCopy^:=Notify;
   fNotifications.Add(LocalCopy);
  end;

 procedure TCustom400.RemoveNotification(Notify:TNotificationEvent);
  var
   i:integer;
   LocalCopy:^TNotificationEvent;
  begin
   for i:=0 to fNotifications.Count-1 do begin
    LocalCopy:=fNotifications[i];
    if @LocalCopy^=@Notify then begin
     Dispose(LocalCopy);
     fNotifications.Delete(i);
     exit;
    end;
   end;
  end;

 procedure TCustom400.DoNotifications(Notify:TNotify);
  var
   i:integer;
  begin
   for i:=0 to fNotifications.Count-1 do
    TNotificationEvent(fNotifications[i]^)(Self,Notify);
  end;

Procedure TCustom400.SetError(Value:Integer);
 begin
  fError:=Value;
  if (fError<>0) then begin
   if Assigned(EOnError) then
    EOnError(Self)
   else
    InternalError;
  end;
 end;

Procedure TCustom400.InternalError;
 Var
  ErrorCount:dword;
  ErrorMsg  :string;
  Buffer    :array[0..100] of char;
  RetLen    :dword;
  i:dword;
 begin
  ErrorMsg:=ClassName+'('+Name+').InternalError #'+IntToStr(Error);
  cwbSV_GetErrCount(ErrHandle, ErrorCount);
  for i:= 1 to  ErrorCount do begin
   if cwbSV_GetErrTextIndexed(ErrHandle, i, Buffer, sizeof(Buffer), RetLen)=0 then
    ErrorMsg:=ErrorMsg+#13#10+Buffer
   else
    ErrorMsg:=ErrorMsg+#13#10+'Error #'+IntToStr(i)+' need '+IntToStr(RetLen)+' bytes.';
  end;
  raise Exception.Create(ErrorMsg);
 end;

{ TAs400 --------------------------------------------------------------- }
Constructor TAs400.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  fDBHCount:=0; fDBHandle:=0;
  fRCHCount:=0; fRCHandle:=0;
 end;

Destructor TAs400.Destroy;
 begin
  if fDBHandle<>0 then begin fDBHCount:=1; DeleteDBHandle; end;
  if fRCHandle<>0 then begin fRCHCount:=1; DeleteRCHandle; end;
  inherited Destroy;
 end;

Procedure TAs400.Loaded;
 begin
  inherited Loaded;
  DoNotifications(tnLoaded);
 end;

Procedure TAS400.CreateDBHandle;
 begin
  inc(fDBHCount);
 end;

Procedure TAS400.DeleteDBHandle;
 begin
  if fDBHCount>0 then dec(fDBHCount);
  if (fDBHCount=0)and(fDBHandle<>0) then begin
   Error:=cwbDB_StopServer(fDBHandle, ErrHandle);
   Error:=cwbDB_DeleteConnectionHandle(fDBHandle, ErrHandle);
   fDBHandle:=0;
  end;
 end;

Function TAS400.DBHandle:THandle;
 begin
  if fDBHandle=0 then begin
   Error:=cwbDB_CreateConnectionHandle(pchar(fSysused), fDBHandle, ErrHandle);
   Error:=cwbDB_StartServer(fDBHandle, ErrHandle);
  end;
  Result:=fDBHandle;
 end;

Procedure TAS400.CreateRCHandle;
 begin
  inc(fRCHCount);
 end;

Procedure TAS400.DeleteRCHandle;
 begin
  if fRCHCount>0 then dec(fRCHCount);
  if (fRCHCount=0)and(fRCHandle<>0) then begin
   Error:=cwbRC_StopSys(fRCHandle);
   fRCHandle:=0;
  end;
 end;

Function TAS400.RCHandle:THandle;
 begin
  if fRCHandle=0 then Error:=cwbRC_StartSys(pchar(fSysUsed),'VCL400',fRCHandle);
  Result:=fRCHandle;
 end;

Procedure TAS400.GetLibList(List:TStrings);
 var
  Request:THandle;
  Data:THandle;
  Indicator:THandle;
  Format:THandle;
  DataSize:dword;
  RowSize:dword;
  RowCount:dword;
  Buffer:PChar;
  i,j:dword;
  s:string;
 begin
  List.Clear;
  try
  CreateDBHandle;
  screen.cursor:=crHourglass;
  // beta version... probably a new TCatalog Object in futur version...
  Error:=cwbDB_CreateCatalogRequestHandle(DBHandle, Request, ErrHandle);
  Error:=cwbDB_RetrieveLibraryInformation(Request,CWBDB_GET_LIBRARY_NAME,ErrHandle);

  Error:=cwbDB_CreateDataHandle(Data, ErrHandle);
  Error:=cwbDB_CreateDataHandle(Indicator, ErrHandle); // NULL, ERROR, DATA
  Error:=cwbDB_CreateDataFormatHandle(DBHandle, Format, ErrHandle);

  Error:=cwbDB_ReturnData(Request, Data, Indicator, Format, ErrHandle);
  Error:=cwbDB_GetData(Request, ErrHandle);

  Error:=cwbDB_GetDataLength(Data, DataSize, ErrHandle);
  if DataSize>0 then begin
   Error:=cwbDB_GetRowSize(Format, CWBDB_LOCAL, RowSize, ErrHandle);
   RowCount:=DataSize div RowSize;
   Error:=cwbDB_GetDataPointer(Data, pointer(Buffer), ErrHandle);
   for i:=1 to RowCount do begin
    s:=Buffer;
    j:=Rowsize;
    while (j>0)and(s[j]=' ') do dec(j);
    SetLength(s,j);
    List.Add(s);
    Buffer:=@Buffer[RowSize];
   end;
  end;

  Error:=cwbDB_DeleteDataFormatHandle(Format, ErrHandle);
  Error:=cwbDB_DeleteDataHandle(Indicator, ErrHandle);
  Error:=cwbDB_DeleteDataHandle(Data, ErrHandle);

  Error:=cwbDB_DeleteCatalogRequestHandle(Request, ErrHandle);
  finally
   DeleteDBHandle;
   screen.cursor:=crDefault;
  end;
 end;

Procedure TAS400.GetLibFiles(LibName:string; FileSelection,FileAttributes:dword; List:TStrings);
 var
  Request:THandle;
  Data:THandle;
  Indicator:THandle;
  Format:THandle;
  DataSize:dword;
  RowSize:dword;
  RowCount:dword;
  Buffer:PChar;
  i,j:dword;
  s:string;
 begin
  List.Clear;
//  if not Active then exit;
  try
  CreateDBHandle;
  screen.cursor:=crHourglass;
  // beta version... probably a new TCatalog Object in futur version...
  Error:=cwbDB_CreateCatalogRequestHandle(DBHandle, Request, ErrHandle);
  Error:=cwbDB_SetLibraryName(Request, pchar(LibName), ErrHandle);
  Error:=cwbDB_SetFileAttributes(Request,FileAttributes,ErrHandle);
// don't work ?
  Error:=cwbDB_SetFileType(Request,FileSelection,ErrHandle);
  Error:=cwbDB_RetrieveFileInformation(Request,CWBDB_GET_FILE_NAME,ErrHandle);

  Error:=cwbDB_CreateDataHandle(Data, ErrHandle);
  Error:=cwbDB_CreateDataHandle(Indicator, ErrHandle); // NULL, ERROR, DATA
  Error:=cwbDB_CreateDataFormatHandle(DBHandle, Format, ErrHandle);

  Error:=cwbDB_ReturnData(Request, Data, Indicator, Format, ErrHandle);
  Error:=cwbDB_GetData(Request, ErrHandle);

  Error:=cwbDB_GetDataLength(Data, DataSize, ErrHandle);
  if DataSize>0 then begin
   Error:=cwbDB_GetRowSize(Format, CWBDB_LOCAL, RowSize, ErrHandle);
   RowCount:=DataSize div RowSize;
   Error:=cwbDB_GetDataPointer(Data, pointer(Buffer), ErrHandle);
   for i:=1 to RowCount do begin
    s:=Buffer;
    j:=Rowsize;
    while (j>0)and(s[j]=' ') do dec(j);
    SetLength(s,j);
    List.Add(s);
    Buffer:=@Buffer[RowSize];
   end;
  end;

  Error:=cwbDB_DeleteDataFormatHandle(Format, ErrHandle);
  Error:=cwbDB_DeleteDataHandle(Indicator, ErrHandle);
  Error:=cwbDB_DeleteDataHandle(Data, ErrHandle);

  Error:=cwbDB_DeleteCatalogRequestHandle(Request, ErrHandle);
  finally
   DeleteDBHandle;
   screen.cursor:=crDefault;
  end;
 end;

Procedure TAs400.SetSysName(Value:string);
 var
  needed:dword;
 begin
  if fSysName=value then exit;
  DoNotifications(tnClose);
  fSysName:=Value;
  if UpperCase(Value)='*DEFAULT' then begin
   SetLength(fSysUsed,10);
   Error:=cwbCO_GetDefaultSysName(pchar(fSysUsed),10,needed,ErrHandle);
   SetLength(fSysUsed,needed);
  end else
   fSysUsed:=Value;
  DoNotifications(tnOpen);
 end;

Procedure TAs400.SetNoWrite(Value:string);
 begin
  if csDesigning in ComponentState then raise exception.Create('this is a read-only property');
 end;

Function TAs400.GetUserName:string;
 var
  size:dword;
 begin
  SetLength(Result,10); Size:=10;
  Error:=cwbCO_GetUserID(pchar(fSysUsed),pchar(Result),CWBCO_CURRENT_USER_ID,Size);
  SetLength(Result,Size);
 end;

Function TAs400.GetVersion:string;
 var
  ver,rel:dword;
 begin
  Error:=cwbCO_GetHostVersion(pchar(fSysUsed),ver,rel);
  result:=Format('V%dR%d',[ver,rel]);
 end;

// TSysList
Procedure TSysList.loaded;
 begin
  inherited Loaded;
  Style:=csDropDownList;
  GetSysList(Items);
  UpdateAS400;
 end;

procedure TSysList.notification(AComponent: TComponent; Operation: TOperation);
 begin
  inherited notification(AComponent,Operation);
  if (AComponent=fAS400)and(Operation=opRemove) then AS400:=nil;
 end;

 procedure TSysList.SetAutoSelect(Value:boolean);
  begin
   fAutoSelect:=Value;
   UpdateAs400;
  end;

 procedure TSysList.SetAS400(Value:TAS400);
  begin
   if fAS400=Value then exit;
   fAS400:=Value;
   UpdateAS400;
  end;

 procedure TSysList.setsysname(Value:string);
  var
   i:integer;
  begin
   i:=Items.IndexOf(Value);
   if i<0 then raise exception.create(cwbUnknowSystem);
   ItemIndex:=i;
   Change;
  end;

 function TSysList.getsysname:string;
  begin
   if ItemIndex<0 then result:='' else result:=Items[ItemIndex];
  end;

 procedure TSysList.Change;
  begin
   UpdateAS400;
   inherited Change;
  end;

 procedure TSysList.UpdateAS400;
  begin
   if (csLoading in ComponentState) then exit;
   if fAS400=nil then exit;
   if fAutoSelect and (ItemIndex<0) and (Items.Count>0) then ItemIndex:=0;
   if ItemIndex<0 then exit;
   if fAS400.Sysname=Text then exit;
   fAS400.Sysname:=Text;
  end;

// TLibFiles
Constructor TLibFiles.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  fActive:=true;
 end;

Destructor TLibFiles.Destroy;
 begin
  AS400:=nil; // RemoveNotification !
  inherited Destroy;
 end;

procedure TLibFiles.SetAS400(Value:TAS400);
 begin
  if value=fAS400 then exit;
  if fAS400<>nil then fAS400.RemoveNotification(AS400Notify);
  fAS400:=Value;
  if fAS400<>Nil then fAS400.AddNotification(AS400Notify);
 end;

procedure TLibFiles.SetLibName(Value:string);
 begin
  fLibName:=Value;
  reload;
 end;

procedure TLibFiles.SetSelectFiles(Value:TSelectFiles);
 begin
  fSelectFiles:=Value;
  reload;
 end;

procedure TLibFiles.SetFileAttributes(Value:TFileAttributes);
 begin
  fFileAttributes:=Value;
  reload;
 end;

procedure TLibFiles.Reload;
 begin
  if (Active)and(fAS400<>Nil)and(not (csLoading in fAS400.ComponentState)) then
   fAS400.GetLibFiles(fLibName,Ord(SelectFiles)+1,Ord(FileAttributes)+1,Items);
 end;

procedure TlibFiles.AS400Notify(Sender:TObject; Notify:TNotify);
 begin
  case Notify of
   tnOpen    : AS400.GetLibFiles(fLibName,Ord(SelectFiles)+1,Ord(FileAttributes)+1,Items);
   tnClose   : Items.Clear;
   tnDestroy : AS400:=nil;
  end;
 end;

 //  TLibFileSelection

 procedure TLibFileSelection.Loaded;
  begin
   inherited Loaded;
   Style:=csDropDownList;
   items.add('All files');
   items.add('Data (PF-DTA)');
   items.add('Logical (LF)');
   items.add('Physical (PF)');
   items.add('Source (PF-SRC)');
   ItemIndex:=0;
  end;

 procedure TLibFileSelection.UpdateLibFiles;
  begin
   if fLibFiles=nil then exit;
   case ItemIndex of
    0 : { All Files       } with fLibFiles do begin SelectFiles:=sfAll;    FileAttributes:=faAll end;
    1 : { Data (PF-DTA)   } with fLibFiles do begin SelectFiles:=sfData;   FileAttributes:=faAll end;
    2 : { Logical (LF)    } with fLibFiles do begin SelectFiles:=sfData;   FileAttributes:=faLogical end;
    3 : { Physical (PF)   } with fLibFiles do begin SelectFiles:=sfData;   FileAttributes:=faPhysical end;
    4 : { Source (PF-SRC) } with fLibFiles do begin SelectFiles:=sfSource; FileAttributes:=faPhysical end;
   end;
  end;

 procedure TLibFileSelection.SetLibFiles(Value:TLibFiles);
  begin
   fLibFiles:=Value;
   UpdateLibFiles;
  end;

 procedure TLibFileSelection.Change;
  begin
   inherited Change;
   UpdateLibFiles;
  end;

 procedure TLibFileSelection.Notification(AComponent: TComponent; Operation: TOperation);
  begin
   inherited Notification(AComponent,Operation);
   if (AComponent=fLibFiles)and(Operation=opRemove) then LibFiles:=nil;
  end;

// TLibList
Destructor TLibList.Destroy;
 begin
  AS400:=nil; // RemoveNotification !
  inherited Destroy;
 end;

procedure TLibList.Notification(AComponent: TComponent; Operation: TOperation);
 begin
  inherited Notification(AComponent,Operation);
  if (AComponent=fLibFiles) and (Operation=opRemove) then fLibFiles:=nil;
 end;

procedure TLibList.Click;
 begin
  if (fLibFiles<>nil)and(ItemIndex>=0) then fLibFiles.LibName:=Items[ItemIndex];
  inherited Click;
 end;

procedure TLibList.SetAS400(Value:TAS400);
 begin
  if value=fAS400 then exit;
  if fAS400<>nil then fAS400.RemoveNotification(AS400Notify);
  fAS400:=Value;
  if fAS400<>Nil then fAS400.AddNotification(AS400Notify);
 end;

procedure TLibList.SetLibFiles(Value:TLibFiles);
 begin
  if value=fLibFiles then exit;
  fLibFiles:=Value;
  fLibFiles.AS400:=fAS400;
  fLibFiles.LibName:=Text;
 end;

procedure TLibList.SetAutoSelect(Value:boolean);
 begin
  fAutoSelect:=Value;
  if ItemIndex<0 then begin ItemIndex:=0; Click end;
 end;

procedure TlibList.AS400Notify(Sender:TObject; Notify:TNotify);
 begin
  case Notify of
   tnOpen    : begin AS400.GetLibList(Items); AutoSelect:=AutoSelect end;
   tnClose   : Items.Clear;
   tnDestroy : AS400:=nil;
  end;
 end;

//----------------- some usefull functions ------------------//

Procedure GetSysList(List:TStrings);
 const
  syssize=10;
 Var
  ListHandle:THandle;
  ListCount :dword;
  system    :dword;
  sysname   :array[0..syssize] of char;
  needsize  :dword;
  error     :integer;
 begin
  List.Clear;
  error:=cwbCO_CreateSysListHandle(ListHandle,ErrHandle);
  if error=0 then begin
   error:=cwbCO_GetSysListSize(ListHandle,ListCount);
   if error=0 then begin
    for system:=1 to ListCount do begin
     cwbCO_GetNextSysName(ListHandle,sysname,syssize,needsize);
     List.Add(sysname);
    end;
   end;
   cwbCO_DeleteSysListHandle(ListHandle);
  end;
 end;

// TParm400
Constructor TParm400.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  fCall400:=nil;
  fSize:=0;
 end;

Destructor TParm400.Destroy;
 begin
  Size:=0;
  if fCall400<>nil then fCall400.fParms.Remove(Self);
  inherited Destroy;
 end;

function TParm400.HasParent: Boolean;
 begin
  Result:=True; // "override" Parent
 end;

function TParm400.GetParentComponent: TComponent;
 begin
  Result:=fCall400; // "new" Parent
 end;

procedure TParm400.SetParentComponent(AParent: TComponent);
 begin
  if fCall400<>nil then fCall400.fParms.Remove(Self);
  fCall400:=AParent as TCall400;
  if fCall400<>nil then fCall400.fParms.add(Self);
 end;

procedure TParm400.ReadState(Reader: TReader);
begin
  inherited ReadState(Reader);
  if Reader.Parent is TCall400 then fCall400 := TCall400(Reader.Parent);
end;

Procedure TParm400.SetSize(Value:word);
 begin
  if fSize>0 then FreeMem(fBind,fSize);
  fSize:=Value;
  if fSize>0 then begin GetMem(fBind,fSize); Clear end;
 end;

Procedure TParm400.Clear;
 begin
  Case ParmType of
   tChar: FillChar(fBind^,Size,' ');
   else   FillChar(fBind^,size,0);
  end;
 end;

Function TParm400.GetString:string;
 begin
  Case ParmType of
   tChar:Result:=pchar(Bind);
  end;
 end;

Procedure TParm400.SetString(Value:string);
 begin
  Case ParmType of
   tChar: StrLCopy(Pchar(fBind),PChar(Value),Size);
  end;
 end;

Procedure TParm400.Add;
 begin
  cwbRC_AddParm(fCall400.fPgmHandle,CWBRC_INOUT {or CWBRC_TEXT_CONVERT},fSize,fBind);
 end;

procedure TParm400.EbcdicToAnsi;
 begin
  case ParmType of
   tChar:EBCDIC.EbcdicToAnsi(fBind^,Size);
  end;
 end;

procedure TParm400.AnsiToEbcdic;
 begin
  case ParmType of
   tChar:EBCDIC.AnsiToEbcdic(fBind^,Size);
  end;
 end;

// TCall400

Constructor TCall400.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  fParms:=TList.Create;
 end;

Destructor TCall400.Destroy;
 begin
  Active:=False;
  While ParmCount>0 do Parms[0].Free;
  fParms.Free;
  inherited Destroy;
 end;

{$IFDEF VER90}
procedure TCall400.GetChildren(Proc: TGetChildProc);
{$ELSE}
procedure TCall400.GetChildren(Proc: TGetChildProc;Root:TComponent);
{$ENDIF}
var
  i:Integer;
  Parm400:TParm400;
begin
  for i := 0 to ParmCount - 1 do  begin
    Parm400:=Parms[i];
    if Parm400.Owner <> Self then Proc(Parm400);
  end;
end;

Procedure TCall400.SetActive(Value:boolean);
 begin
  if fActive=Value then exit;
  if csLoading in ComponentState then
   fActivate:=Value
  else begin
   if Value then Activate else begin
    fActive:=False;
    Error:=cwbRC_DeletePgm(fPgmHandle);
   end;
  end;
 end;

Procedure TCall400.SetAS400(Value:TAS400);
 begin
  if fAS400<>nil then begin
   Active:=False;
   fAS400.RemoveNotification(Notify400);
   fAS400.DeleteRCHandle;
  end;
  fAS400:=Value;
  if fAS400<>nil then begin
   fAS400.AddNotification(Notify400);
   fAS400.CreateRCHandle;
  end;
 end;

procedure TCall400.Activate;
 var
  i:integer;
 begin
  try
   fActive:=false;
   screen.cursor:=crHourglass;
   Error:=cwbRC_CreatePgm(pchar(fPgmName),pchar(fLibName),fPgmHandle);
   For i:=0 to ParmCount-1 do Parms[i].Add;
   fActive:=True;
  finally
   screen.cursor:=crDefault;
  end;
 end;

procedure TCall400.Notify400(Sender:TObject; Notify:TNotify);
 begin
  case Notify of
   tnLoaded  : fActive:=fActivate;
   tnClose   : Active:=False;
   tnDestroy : AS400:=nil;
  end;
 end;

function TCall400.GetParm(Index:Integer):TParm400;
 begin
  Result:=fParms[Index];
 end;

function TCall400.GetParmCount:integer;
 begin
  Result:=fParms.Count;
 end;

function TCall400.ParmByName(Const ParmName:string):TParm400;
 var
  i:integer;
 begin
  for i:=0 to ParmCount-1 do begin
   Result:=fParms[i];
   if Result.Name=ParmName then exit;
  end;
  Result:=nil;
 end;

Function TCall400.AddParm(AName:string; AType:TParmType; ASize,APrecision:word):pointer;
 begin
  if Active then raise Exception.Create(cwbNotWhileActive);
  with TParm400.Create(Owner) do begin
   Name:=AName;
   SetParentComponent(Self);
   ParmType:=AType;
   Size:=ASize;
   Precision:=APrecision;
   Result:=fBind;
//   Clear;
  end;
 end;

Procedure TCall400.AssignParms(ACall400:TCall400);
 var
  i:integer;
  Parm:TParm400;
 begin
  if Active then raise Exception.Create(cwbNotWhileActive);
 // remove current parms
  While ParmCount>0 do Parms[0].Free;
 // copy parms
  for i:=0 to ACall400.ParmCount-1 do begin
   Parm:=ACall400.Parms[i];
   AddParm(Parm.Name,Parm.ParmType,Parm.Size,Parm.Precision);
  end;
 end;


Function TCall400.Execute:boolean;
 var
  i:integer;
 begin
  Result:=False;
  try
   Active:=True;
   screen.cursor:=crHourglass;
   for i:=0 to ParmCount-1 do Parms[i].AnsiToEbcdic;
   Error:=cwbRC_CallPgm(fAS400.RCHandle,fPgmHandle,ErrHandle);
   Result:=True;
  finally
   for i:=0 to ParmCount-1 do Parms[i].EbcdicToAnsi;
   Screen.cursor:=crDefault;
  end;
 end;

initialization
 // create an Error Handle
 if cwbSV_CreateErrHandle(ErrHandle)<>0 then begin
  MessageBox(0,cwbNoErrHandle,cwbTitle,MB_ICONWARNING);
 end;
 ErrorMonitor:=TErrorMonitor.Create(nil);

finalization
 // delete Error Handle
 ErrorMonitor.Free;
 cwbSV_DeleteErrHandle(ErrHandle);

end.
